/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


	/*
	 * a library of plotting routines that may be called by
	 * other programs.  All external functions have names
	 * prefixed by "xplot_"
	 *
	 * This file is organized in three pieces:
	 *
	 *    1. mostly functions that are externally visible and
	 *       translate the user's request into internal format
	 *    2. a set of math and Babylonian format conversions
	 *    3. functions that actually call X Windows routines, 
	 *       only a few of which are visible to the user.
	 *
 	 *  MWR 8/18/96
	 *
	 * 
	 * Made minor changes to 'xplot_run_command' so that is skips
	 * comment lines, and completely blank lines, in command files.
	 *  MWR 2/21/97
	 */

/*
 * CHANGES:
 *
 * 8/19/96  - added a call to XFillRectangle in initXstuff, so that
 *              the new graphics window is erased before being displayed.  MWR
 *
 * 3/7/2003 - removed (re)-declaration of "strdup" function in the install()
 *              function.  It caused parse errors, and isn't necessary.  MWR
 */
	 

#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

#include "xplotlib.h"
#include "adobe.h"
#include "usleep.h"



typedef short INT16;
typedef int (*PFI)();

#define 	fabs(x)		(((x) < 0.) ? -(x): (x))

	/* 
	 * some #defininitions that aren't of interest to the
	 * outside user of the XPLOT routines
	 */
	/* max length of input lines */
#define LINELEN        100
	/* max number of commands we can install in hash table */
#define MAXCOM         100
	/* least number of letters to make a command via pattern-matching */
#define MIN_LETTERS      1		
	/* stuff used by the 'help' command */
#define	HELPCHAR        '&'
#define LINESIZE       100
#define TEXT_HEIGHT     25           /* number of lines printed in each */
                                     /*   installment of a help message */
	/* types of labelling of the axes */
#define MODE_NUMERIC     0 
#define MODE_BABY        1

#ifndef M_PI
#define M_PI 3.14159265359
#endif

	/* size of display windows */
#define GRAPHICS_HEIGHT   430		/* in pixels */
#define GRAPHICS_WIDTH    500		/* in pixels */
	/* initial places on the screen */
#define TEXT_XOFFSET      20
#define TEXT_YOFFSET       0
#define GRAPHICS_XOFFSET 550
#define GRAPHICS_YOFFSET  50


#define AR_SIZE 50000

#define HASHMARKSIZE  5		/* a good size for hash marks on axes */

	/* point stuff */
#define P_MAXVERT    20
#define P_SIZE        6		/* a good scaling factor for size */

	/* 
	 * this many lines are printed before user is prompted to
	 * continue or quit.  
	 */
#define HELP_LINES   20


	/* 
	 * we _shouldn't_ need to do this on the Sun, but with
	 * SunOS 5.5 and gcc 2.7.2.f.1, we _do_ need to do so.
	 */
#ifdef SUN
extern int kill(pid_t proc_id, int kill_signal);
#endif


	/* the commands are placed into a hash table of these structures */
struct nlist {
	char *name;
	PFI func;
	int match;
};
	

	/* static functions which shouldn't be visible to user */

/* returns 1 if valid babylonian string */ 
static int is_hms(char *in);
/* converts double to babylonian notation */
static char *d2hms(double x, int mode);	
/* converts babylonian string to double */
static double hms2d(char *in);				

static void install(char *name, PFI func);
static void box(void);
static void xaxis(double first, double dx);
static void yaxis(double first, double dx);
static int inside(double x, double y);
static int code(double x, double y);
static void setptype(double in);
static void normalize(void);
static void clip_n_draw(double x1, double y1, double x2, double y2);
static void show_help(FILE *fin);
static void draw_marker(INT16 x, INT16 y);
static void do_flush(void);
static void world_to_ndc(double wx, double wy, double *nx, double *ny);
static void ndc_to_world(double nx, double ny, double *wx, double *wy);
static void find_digits(double lo, double hi, char *format);
static int minmatch(char *);
static struct nlist *lookup(char *);
static int my_get_col(char *, int, char **);
static void die(char *);
static void rationalize(double bottom, double top, int *nticks, 
                       double *first, double *delta);
static int iscomment(char *in);
#if 0
static double e_to_f(double x, int ex);
static int f_to_e(double f, double *frac, int *ex);
#endif


	/* declarations for X Windows routines herein */

static void initXstuff(void);
static void endXstuff(void);
static void savepixmaps(void);
static int get_char_width(void);
static void draw_poly(int n, INT16 *pts);
static void draw_text(INT16 x, INT16 y, char *s);
static void get_norm_GC(Window window, GC *gc, XFontStruct *fontinfo);
static void get_inverted_GC(Window window, GC *gc, 
                            XFontStruct *fontinfo);
static void localcursor(int x, int y);



	/* variables static to this file */

static int MINX, MAXX, MINY, MAXY;
static int T_HEIGHT, T_WIDTH;
static int cur_p_type, cur_l_type, cur_p_fill;
static int high_line, low_line;
static int xc, yc;
static int ec, pc;
static double x_text_pos, y_text_pos;		/* pos where text to be drawn next */
static double p_size;
static double file_x[AR_SIZE], file_y[AR_SIZE];	/* raw coords from file */
static double x[AR_SIZE], y[AR_SIZE];		/* same, but translated to NDC */
static double file_e[AR_SIZE], file_p[AR_SIZE];	/* added rrt */
static double minx, maxx, miny, maxy;
static int num_read;
static FILE *savefile;
static short SAVE;
static short PRINT;
static int x_mode=MODE_NUMERIC;
static int y_mode=MODE_NUMERIC;
static int child_id;
static int ncom;
static struct nlist com[MAXCOM];
static FILE *fp;
static char my_filename[80];
static int xplot_status_flag = XPLOT_OK;

   /* 
    * these variables are set to world coords whenever cursor button
    * is pressed, and they can be read by the "xplot_get_cursor()" function
    */
static double cursor_xpos;       
static double cursor_ypos;




	/* 
	 * get rid of the windows and all associated trappings.
	 * set the static variable xplot_status_flag to XPLOT_QUIT, signalling 
	 * xplot_command to return XPLOT_QUIT.
	 */

int
xplot_quit(void)
{

	adobe_cleanup();
	endXstuff();
	printf("\r");
	kill(child_id, SIGTERM);

	/* this signals that the user called xplot_quit() */
	xplot_status_flag = XPLOT_QUIT;

	return(0);
}


	/* 
	 * kill the parent process, but leave the graphics window
	 * sitting in the display.  The user will have to kill it
	 * with his window manager.
	 */

int
xplot_spawn(void)
{
	char s[100];

	if (fork() == 0) {
		while (1) {
			xplot_event(s, XPLOT_EVENT_LOOP);
		}
	}

	/* 
	 * now do the same stuff as "quit", but 
	 * don't kill graphics window 
	 */
	adobe_cleanup();
	printf("\r");
	kill(child_id, SIGTERM);
	exit(0);
}
	

void
xplot_init(void)
{
	initXstuff();

	T_WIDTH = get_char_width();
	T_HEIGHT = T_WIDTH;
	MINX = 10 * T_WIDTH;
	MAXX = GRAPHICS_WIDTH - 6 * T_WIDTH;
	MAXY = 5 * T_HEIGHT;
	MINY = GRAPHICS_HEIGHT - 7 * T_HEIGHT;

	/* set some characteristics of the graphics window */
	p_size = 1.0;
	xplot_ptype("ptype 4");
	cur_p_fill = 0;		/* polygons not filled in */
	cur_l_type = SOLID;

	/* set a few other things */
	low_line = 1;
	high_line = AR_SIZE;
	minx = miny = 0;
	maxx = maxy = 1;

	/* initialize flags */
	PRINT = 0;
	SAVE = 0;
	savefile = NULL;

	/* install commands in the command table */
	install("clear", xplot_clear);
	install("erase", xplot_clear);
	install("data", xplot_get_file);
	install("xc", xplot_get_xc);
	install("yc", xplot_get_yc);
	install("xlog", xplot_x_log);
	install("ylog", xplot_y_log);
	install("xdelog", xplot_x_delog);
	install("ydelog", xplot_y_delog);
	install("hardcopy", xplot_do_hardcopy);
	install("lines", xplot_lines);
	install("limits", xplot_get_limits);
	install("ltype", xplot_ltype);
	install("ptype", xplot_ptype);
	install("expand", xplot_expand);
	install("points", xplot_points);
	install("connect", xplot_connect);
	install("hist", xplot_hist);	
	install("dot", xplot_dot);
	install("draw", xplot_draw);
	install("relocate", xplot_position);
	install("position", xplot_position);
	install("cursor", xplot_cursor);
	install("label", xplot_label);
	install("reverse", xplot_reverse);
	install("quit", xplot_quit);
	install("end", xplot_quit);
	install("exit", xplot_quit);
	install("spawn", xplot_spawn);
	install("box", xplot_draw_axes);
	install("xlabel", xplot_xlabel);
	install("ylabel", xplot_ylabel);
	install("title", xplot_title);
	install("help", xplot_help);
	install("save", xplot_save_com);
	install("nosave", xplot_stop_save);
	install("run", xplot_run_cmd);
	install("input", xplot_run_cmd);
	install("plot", xplot_plot);

	install("vfield", xplot_vfield);	
	install("errorbar", xplot_errorbar);
	install("ec", xplot_get_ec);		
	install("pc", xplot_get_pc);	
	install("wait", xplot_wait);	
	install("xsixty", xplot_get_xsixty);	
	install("ysixty", xplot_get_ysixty);	

}


	/*
	 * given some string containing an XPLOT command and any
	 * arguments it requires, look up the command in the hash table
	 * and attempt to execute it.  
	 *
	 * We return the value of the static variable "xplot_status_flag",
	 * which can be 
	 *
	 *       XPLOT_OK               all is well, let's keep going
	 *       XPLOT_ERROR            something bad happened in last command
	 *       XPLOT_QUIT             we've been asked to quit
	 */

int
xplot_command(char *string)
{
	struct nlist *np;

	if ((np = lookup(string)) == NULL) {
		printf("unknown command %s\n", string);
		return(XPLOT_ERROR);
	}
	(*(np->func))(string);
	if ((SAVE) && (strcmp(string, "save") != 0)) {
		fprintf(savefile, "%s\n", string);
	}
	do_flush();
	savepixmaps();
	return(xplot_status_flag);
}



static void
die(s)
char *s;
{
	adobe_cleanup();
	endXstuff();
	printf("\r");
	kill(child_id, SIGTERM);
	fprintf(stderr, "%s\n", s);
	exit(-1);
}

	/* run commands from a file */

int
xplot_run_cmd(cmd_buf)
char *cmd_buf;
{
	char dummy[10], file[10];
	FILE *cmd_file;
	long floop=0;
	int i, j;
	char line[LINELEN], command[10];
	struct nlist *np;
	int errcode;

	if (sscanf(cmd_buf, "%s %s", dummy, file) != 2) {
		printf("usage: run file\n");
		return(1);
	}
	if ((cmd_file = fopen(file, "r")) == NULL) {
		printf("run: error in opening file %s\n", file);
		return(1);
	}
	printf("executing file %s\n", file);
	for (i = 1; fgets(line, LINELEN, cmd_file) != NULL; i++) {
		line[strlen(line)-1] = ' ';
		command[0] = '\0';
		sscanf(line, "%s", command);

		/* ignore comment lines */
                if (iscomment(command)) {
			continue;
		}

		/* ignore completely blank lines */
		for (j = 0; j < strlen(command); j++) {
			if (!isspace(command[j])) {
				break;
			}
		}
		if (j == strlen(command)) {
			continue;
		}

		if(strcmp(command ,"loop") == 0){
			int rval;
			rval=fseek(cmd_file,floop,SEEK_SET);
			printf("looping %d\n",rval);
			continue;
		}
		if(strcmp(command ,"begin") == 0){
			floop=ftell(cmd_file);
			continue;
		}
		if ((np = lookup(command)) == NULL) {
			printf("line %d: syntax error\n", i);
			break;
		}
		errcode = (*(np->func))(line);
		if (errcode != 0) {
			printf("line %d: error\n", i);
			break;
		}
	}
	fclose (cmd_file);
	return 0;
}


	/* in the second two arguments, put the NDC coordinates corresponding 
	   to the world coordinates given in the first two arguments */
	/* this looks wrong because I had to change it for the RT - which
	   has the (0,0) coordinate in the UPPER-left, not LOWER-left, 
	   corner */

static void
world_to_ndc(wx, wy, nx, ny)
double wx, wy, *nx, *ny;
{
	double fx, fy, cx, cy;

	fx = (MAXX - MINX) / (maxx - minx);
	fy = (MINY - MAXY) / (maxy - miny);
	cx = MINX - fx * minx;
	cy = MAXY + fy * maxy;

	*nx = wx * fx + cx;
	*ny = cy - wy * fy;

}

	/* in the second two arguments, put the world coordinates corresponding
	   to the NDC coordinates given in the first two arguments.
	   again, looks wierd due to IMB convention of Y having 0 at top of
	   screen. */

static void
ndc_to_world(nx, ny, wx, wy)
double nx, ny, *wx, *wy;
{
	double fx, fy, cx, cy;

	fx = (MAXX - MINX) / (maxx - minx);
	fy = (MINY - MAXY) / (maxy - miny);
	cx = MINX - fx * minx;
	cy = MAXY + fy * maxy;

	*wx = (nx - cx) / fx;
	*wy = (cy - ny) / fy;
}


	/* set the type of line used in do_connect(); with no arguments,
	   prints the current type */

int
xplot_ltype(cmd_buf)
char *cmd_buf;
{
	int type;
	char dummy[7];

	if (sscanf(cmd_buf, "%s %d", dummy, &type) < 2) {
		printf("%d\n", cur_l_type);
	}
	else {
		if ((type >= MIN_L_TYPE) && (type <= MAX_L_TYPE))
			cur_l_type = type;
	}
	adobe_setdash(cur_l_type);
	return(0);
}

	/* set the size of the point markers used by 'points' command 
	   without args, print the current value */

int
xplot_expand(cmd_buf)
char *cmd_buf;
{
	double size;
	char dummy[9], foo[20];

	if (sscanf(cmd_buf, "%s %lf", dummy, &size) < 2) {
		printf("%f\n", p_size);
	}
	else {
		p_size = size;
		sprintf(foo, "ptype %d", cur_p_type);
		xplot_ptype(foo);
	}
	return(0);
}


/*	print the stored record
*/
int
xplot_do_hardcopy(cmd_buf)
char *cmd_buf;
{
	char dummy[10],other[10];

	if(sscanf (cmd_buf, "%s %s", dummy, other) != 2)
		strcpy(other,"");
	adobe_close(other);

	return 0;
}



	/* draw the axes of the graph: if arguments are given, use them to
		scale the axes, otherwise, figure out appropriate scales */

int
xplot_draw_axes(cmd_buf)
char *cmd_buf;
{
	char dummy[10];
	double dx, dy;
	int nx,ny;
	double xfirst,yfirst, xtemp, ytemp;
	INT16 old_style;

	if ((minx == maxx) || (miny == maxy)) {
		printf("axes: graph limits not set\n");
		return(1);
	}

	/* draw a box around the graph */
	box();

	/* set line style to solid and save old line style */
	old_style = cur_l_type;
	cur_l_type = SOLID;

	/* read scale interval from the command line */
	rationalize(minx, maxx, &nx, &xfirst, &dx);
	rationalize(miny, maxy, &ny, &yfirst, &dy);

	if (sscanf (cmd_buf, "%s %lf %lf", dummy, &xtemp, &ytemp) == 3) {
		/* since no intervals are given, determine suitable intervals */
		dx = xtemp ;
		dy = ytemp;
	}

	xaxis(xfirst, dx);
	yaxis(yfirst, dy);

	/* restore default text justification and restore old line style */
	cur_l_type = old_style;

	return(0);
}


	/* draw the hash marks on the x axis and label them; also put the
		scaling factor if there is one */

static void
xaxis(first,dx)
double first,dx;
{

	char dummy[10],format[20];
	double  x, ndcx, ndcy, xn, xx, yn;
	INT16 p[4];


	/* draw hash marks and label them */
	p[1] = MINY + HASHMARKSIZE;
	p[3] = MINY - HASHMARKSIZE;
	
	
	xn = (minx < maxx ? minx : maxx);
	xx = (minx < maxx ? maxx : minx);
	yn = (miny < maxy ? miny : maxy);

	find_digits( xn, xx, format);

	if (minx < maxx) {
		for (x = first; x <= maxx; x += dx) {
			world_to_ndc(x, yn, &ndcx, &ndcy);
			p[0] = ndcx;
			p[2] = ndcx;
			draw_poly(2, p);
			p[0] -= 2*T_WIDTH;
			if(x_mode == MODE_NUMERIC)
				sprintf(dummy, format, x);
			else
				strcpy(dummy,d2hms(x,BABY_MODE_DEC_NO));
			draw_text(p[0], p[1], dummy);
		}
	}
	else {
		for (x = first; x >= maxx; x += dx) {
			world_to_ndc(x, yn, &ndcx, &ndcy);
			p[0] = ndcx;
			p[2] = ndcx;
			draw_poly(2, p);
			p[0] -= 2*T_WIDTH;
			if(x_mode == MODE_NUMERIC)
				sprintf(dummy, format, x);
			else
				strcpy(dummy,d2hms(x,BABY_MODE_DEC_NO));
			draw_text(p[0], p[1], dummy);
		}
	}
}


	/* draw hash marks on the y axis and label them; also put the scaling
		factor if there is one */

static void
yaxis(first,dy)
double first,dy;
{
	char dummy[10],format[10];
	double  y, ndcx, ndcy, xn, yn, yx;
	INT16 p[4], pt;


	/* put hash marks and label them */
	p[0] = MINX - HASHMARKSIZE;
	p[2] = MINX + HASHMARKSIZE;

	xn = (minx < maxx ? minx : maxx);
	yn = (miny < maxy ? miny : maxy);
	yx = (miny < maxy ? maxy : miny);

	find_digits( yn, yx, format);

	if (miny < maxy) {
		for (y = first ; y <= maxy; y += dy) {
			world_to_ndc(xn, y, &ndcx, &ndcy);
			p[1] = ndcy;
			p[3] = ndcy;
			draw_poly(2, p);
			p[1] -= T_HEIGHT;
			if(y_mode == MODE_NUMERIC)
				sprintf(dummy, format, y);
			else
				strcpy(dummy,d2hms(y,BABY_MODE_DEC_NO));
			pt = p[0] - strlen(dummy)*T_WIDTH;
			draw_text(pt, p[1], dummy);
		}
	}
	else {
		for (y = first ; y >= maxy; y += dy) {
			world_to_ndc(xn, y, &ndcx, &ndcy);
			p[1] = ndcy;
			p[3] = ndcy;
			draw_poly(2, p);
			p[1] -= T_HEIGHT;
			if(y_mode == MODE_NUMERIC)
				sprintf(dummy, format, y);
			else
				strcpy(dummy,d2hms(y,BABY_MODE_DEC_NO));
			pt = p[0] - strlen(dummy)*T_WIDTH;
			draw_text(pt, p[1], dummy);
		}
	}
}



	/* label x axis */

int
xplot_xlabel(cmd_buf)
char	*cmd_buf;
{

	INT16 px, py;
	char *p;

	/* skip over the word "xlabel" and blank space to get to the label */
	for (p = cmd_buf ; !(isspace(*p)); p++)
		;
	for ( ; (isspace(*p)); p++)
		;

	/* if no text, just return */
	if (*p == '\0')
		return(0);

	/* position text and print */
	px = MINX + (MAXX - MINX) / 2 - (((int)strlen(p)/2) * T_WIDTH);
	py = MINY + 4 * T_HEIGHT;
	draw_text(px, py, p);

	return(0);
}


	/* label y axis */
	/* make the label like this:   y
	                               l
	                               a
	                               b
	                               e
	                               l
	  */

int
xplot_ylabel(cmd_buf)
char	*cmd_buf;
{

	INT16 px, py, tbit, i;
	char *p, pp[2];

	/* skip over the word "ylabel" and blank space to get to the label */
	for (p = cmd_buf ; !(isspace(*p)); p++)
		;
	for ( ; (isspace(*p)); p++)
		;

	/* if no text, just return */
	if (*p == '\0')
		return(0);

	/* find the place to start the string */
	tbit = T_HEIGHT + 1;
	py = (MINY + MAXY)/2 - ((int)strlen(p)/2) * (T_HEIGHT + tbit);
	px = MINX - 9 * T_WIDTH;
	pp[1] = '\0';

	/* print text */
	for (i = 0; i < (int)strlen(p); i++) {
		pp[0] = p[i];
		draw_text(px, py, pp);
		py += (T_HEIGHT + tbit);
	}

	return(0);
}


	/* put a title on the graph */

int
xplot_title(cmd_buf)
char	*cmd_buf;
{

	INT16 px, py;
	char *p;

	/* skip over the word "title" and any blank space to get to the label */
	for (p = cmd_buf + 5; (isspace(*p)); p++)
		;

	/* if no text, just return */
	if (*p == '\0')
		return(0);

	/* print text */
	px = MINX + (MAXX - MINX) / 2 - ((int)(strlen(p)/2) * T_WIDTH);
	py = MAXY - 3 * T_HEIGHT;
	draw_text(px, py, p);

	return(0);
}


	
/*      makes rational first tic mark value, increment
*/

		
static void 
rationalize(bottom, top, nticks, first, delta)
double bottom, top;     /* Specified virtual plotter limits -- won't exceed!*/
int * nticks;           /* Computed number of ticks -- guarantee 4<=n<=9    */
double * first, * delta;/* Computed first tick position, increment          */
{
    double dx, absdelta, logdelta, newdelta, ntimes, nt, newfirst;

    dx = (top - bottom)/8.0;
    absdelta = fabs( dx );
    if( absdelta == 0.0 ) absdelta = 1.0E-10;
    logdelta = log10( absdelta );
    newdelta = pow( 10.0, floor( logdelta ) );
    while( newdelta < absdelta )
        newdelta = 2.0*newdelta;
    ntimes = floor( bottom / newdelta );
    if( dx < 0.0 )
    {
        newfirst = ntimes*newdelta;
        if( newfirst + newdelta <= bottom ) newfirst += newdelta;
    }
    else
    {
        newfirst = (ntimes + 1.0)*newdelta;
        if( newfirst - newdelta >= bottom ) newfirst -= newdelta;
    }
    nt = floor( fabs(top - newfirst) / newdelta );
    if( newfirst + (1.0 + nt)*newdelta <= top*fabs(dx)/dx ) nt += 1.0;
    * nticks = (int) (nt + 0.5 );
    if( dx < 0.0 ) newdelta = -newdelta;
    * first = newfirst;
    * delta = newdelta;
}

/*	fills out appropriate format statement
	for lower and upper limits
*/

static void
find_digits(lo,hi,format)
double lo,hi;
char *format;
{
	int log_dig;
	double diff;

	diff=fabs(hi-lo);
	if(diff==0.){
		if(hi==0.){
			strcpy(format,"%8.3f");
			return;
		}else
			log_dig=(int)floor(log10(fabs(hi)));	
	}else
		log_dig=(int)floor(log10(diff));	

	if(log_dig<-5){
		strcpy(format,"%.2E");
		return;
	}
	if(log_dig>5){
		strcpy(format,"%.2E");
		return;
	}
	if(log_dig>3){
		strcpy(format,"%.0f");
		return;
	}
	if(log_dig<-3){
		sprintf(format,"%%.%df",0-log_dig);
		return;
	}
	sprintf(format,"%%.%df",3-log_dig);
}
	/* save commands into a file */

int
xplot_save_com(cmd_buf)
char *cmd_buf;
{
	char dummy[10], file[10];

	if (savefile != NULL) {
		printf("save: file already open\n");
		return(1);
	}
	if (sscanf(cmd_buf, "%s %s", dummy, file) != 2) {
		printf("usage: save file\n");
		return(1);
	}
	if ((savefile = fopen(file, "w")) == NULL) {
		printf("save: error in opening file %s\n", file);
		return(1);
	}
	SAVE = 1;
	printf("save start: file is %s\n", file);
	return(0);
}

	/* stop saving commands into a file */

int
xplot_stop_save(cmd_buf)
char *cmd_buf;
{
	if (savefile == NULL) {
		SAVE = 0;
		return(0);
	}
	if (fclose(savefile) == EOF) {
		printf("error in closing file\n");
		SAVE = 0;
		savefile = NULL;
		return(1);
	}
	SAVE = 0;
	savefile = NULL;
	printf("ending save\n");
	return(0);
}


	/* convert a floating point to frac * 10 ^ ex format */

#if 0
static int
f_to_e (f, frac, ex)
double f, *frac;
int *ex;
{

	int sign, i;

	if (f == 0) {
		*frac = 0;
		*ex = 0;
		return(0);
	}

	/* determine the sign and save it for later */
	if (f < 0)
		sign = -1;
	else
		sign = 1;

	/* convert to exponential notation */
	f = fabs(f);
	if (f < 1)
		for (i = 0; f < 1; i--)
			f *= 10;
	else
		for (i = 0; f > 10; i++)
			f /= 10;

	*frac = sign * f;
	*ex = i;

	return(0);
}
#endif	


	/* convert base 10 exponential notation to floating point */

#if 0
static double
e_to_f (x, ex)
double x;
int ex;
{

	if (x == 0)
		return (0);
	if (ex > 0)
		for (; ex > 0; ex--)
			x *= 10;
	else
		for (; ex < 0; ex++)
			x /= 10;
	return (x);
}
#endif

/*	
		function lookup, but
		does NOT actually hash but allows
		least case pattern matching
*/



static void
install(name, func)
char *name;
PFI func;
{

	com[ncom].name=strdup(name);
	com[ncom].func=func;
	com[ncom].match=minmatch(name);
	ncom++;
	if(ncom == MAXCOM)
		die("too many commands installed");
}

struct nlist *
lookup(s)
char *s;
{
	int i;

	for(i=0; i<ncom; i++)
		if(strncmp(s, com[i].name, com[i].match)== 0)
			return &com[i];
	return NULL;
}

/*	computes the minimum number of characters needed to match the
	string
*/


static int 
minmatch(name)
char *name;
{
	int i,try,len,min;

	min=len=strlen(name);

	for(try=MIN_LETTERS; try<min ; try++){
		min=try;
		for(i=0; i<ncom; i++){
			if(strncmp(name,com[i].name,try)==0){
				min=len;
				if(try == com[i].match)
					com[i].match++;
			}
		}
	}
	return min;	
}

#ifdef STRDUP
static char *
strdup(s)
char *s;
{
	char *p;

	if ((p = malloc(strlen(s)+1)) == NULL)
		die("can't allocate string");
	strcpy(p, s);
	return(p);
}	

#endif


int
xplot_help(cmd_buf)
char *cmd_buf;
{
	char command[20], s[20];
	struct nlist *np;
	static FILE *fhelp=NULL;
	char line[LINESIZE+1];

	if( fhelp==NULL){
		if( (fhelp = fopen(HELPFILE, "r")) == NULL) {
			fprintf(stderr, 
				"sorry, cannot open file %s for help\n", HELPFILE);
			return(0);
		}
	}else
		fseek(fhelp,0L,0);		/* rewind the file */

	if (sscanf(cmd_buf, "%s %s", line, command) < 2)
		sprintf(command, "help");

	if((np=lookup(command))!=NULL)		/* invoke least pattern matching */
		strcpy(command,np->name);

	/* try to find the command asked about in the help file */
	while (fgets(line, LINESIZE, fhelp) != NULL) {
		if (line[0] == HELPCHAR) {
			if (sscanf(line+1, "%s", s) < 1)
				continue;
			if (strcmp(s, command) != 0)
				continue;

		/* okay, found the command in the help file - 
		   now, print out the help messages in a nice
		   way. */
			show_help(fhelp);
			return(0);
		}
	}

	/* if we got here, couldn't find help for the given command -
	   so tell the user to say 'help help' next time */
	printf("sorry, no help available for %s\n", command);
	return(0);
}

	/* we're at a point in the help file where the help text for some
	   command begins - we want to print it out
	   until we get to the end of the current help
	   text (which will be marked by a HELPCHAR in the first character
	   of a line from the file). */

static void
show_help(fin)
FILE *fin;
{
	int i;
	char more[LINESIZE+1];
	char line[LINESIZE+1];

	i = 0;
	while(fgets(line, LINESIZE, fin)){
		if(line[0] == HELPCHAR) 
			break;
		if (++i > HELP_LINES) {
			printf(" -- hit return to continue, q to quit -- ");
			fflush(stdout);
			xplot_event(more, XPLOT_EVENT_LOOP);
			if ((more[0] == 'q') || (more[0] == 'Q')) {
				break;
			}
			else {
				i = 0;
			}
		}
		printf("%s",line);	
	}
}

	

	/* made field sizes slightly larger MWR 11/12/1991 */

	/* open the given file, setting global 'fp' to its file pointer */

int
xplot_get_file(cmd_buf)
char *cmd_buf;
{
	char dummy[10];

	sscanf(cmd_buf, "%s %s", dummy, my_filename);
	if( fp != NULL)
		fclose(fp);

	if ((fp = fopen(my_filename, "r")) == NULL) {
		printf("cannot open file %s\n", my_filename);
		return(1);
	}
	return(0);
}
	

	/* grab the x data from the given column.  the 0-th column
	   means to use the line number in the file as the x value
	   (not counting comment lines and/or skipped lines) */

int
xplot_get_xc(cmd_buf)
char *cmd_buf;
{
	int i, k, c1;
	char c[10];
	char buf[BUFSIZ],*buf2;

	if (fp == NULL) {
		printf("no file open\n");
		return(1);
	}
	if (sscanf(cmd_buf, "%s %d", c, &c1) < 2) {
		printf("%d\n", xc);
		return(0);
	}
	xc = c1;

	fseek(fp, (long) 0, 0);
	i = 0;
	k = 1;

	while ((fgets(buf, BUFSIZ, fp) != NULL) && (i < AR_SIZE)) {
		if (k++ < low_line) {
			continue;
		}
		if ((i >= AR_SIZE) || (k > (high_line + 1)))
			break;
		if(iscomment(buf))
			continue;
		if (xc == 0) {
			file_x[i] = (double) i;
		}
		else {
			if(my_get_col(buf,xc,&buf2)){
				printf("column not found on line %d of %s\n", 
					i + 1,my_filename);
				return(1);
			}
			if(sscanf(buf2,"%lf",&file_x[i])!= 1){
				printf("non numeric field on line %d of %s\n", 
					i + 1,my_filename);
				return(1);
			}
		}
		i++;
	}
 	num_read = i;
	x_mode=MODE_NUMERIC;
	return(0);
}


	/* grab the y data from the given column.  See comments above for
	   the meaning of 'yc 0'. */

int
xplot_get_yc(cmd_buf)
char *cmd_buf;
{
	int i, k, c1;
	char c[10];
	char buf[BUFSIZ],*buf2;

	if (fp == NULL) {
		printf("no file open\n");
		return(1);
	}
	if (sscanf(cmd_buf, "%s %d", c, &c1) < 2) {
		printf("%d\n", yc);
		return(0);
	}

	yc = c1;

	fseek(fp, (long) 0, 0);
	i = 0;
	k = 1;

	while ((fgets(buf, BUFSIZ, fp) != NULL) && (i < AR_SIZE)) {
		if (k++ < low_line)
			continue;
		if ((i >= AR_SIZE) || (k > (high_line + 1)))
			break;
		if(iscomment(buf))
			continue;
		if (yc == 0) {
			file_y[i] = (double) i;
		}
		else {
			if(my_get_col(buf,yc,&buf2)){
				printf("inconsistent data on line %d of %s\n", 
					i + 1,my_filename);
				return(1);
			}
			if(sscanf(buf2,"%lf",&file_y[i]) != 1) {
				printf("non numeric field on line %d of %s\n", 
				i + 1,my_filename);
				return(1);
			}
		}
		i++;
	} 	
	num_read = i;
	y_mode=MODE_NUMERIC;
	return(0);
}


int
xplot_get_ec(cmd_buf)
char *cmd_buf;
{
	int i, k, c1;
	char c[10];
	char buf[BUFSIZ],*buf2;

	if (fp == NULL) {
		printf("no file open\n");
		return(1);
	}
	if (sscanf(cmd_buf, "%s %d", c, &c1) < 2) {
		printf("%d\n", ec);
		return(0);
	}
	ec = c1;

	fseek(fp, (long) 0, 0);
	i = 0;
	k = 1;

	while ((fgets(buf, BUFSIZ, fp) != NULL) && (i < AR_SIZE)) {
		if (k++ < low_line)
			continue;
		if ((i >= AR_SIZE) || (k > (high_line + 1)))
			break;
		if(iscomment(buf))
			continue;
		if(my_get_col(buf,ec,&buf2)){
			printf("inconsistent data on line %d of %s\n", 
				i + 1,my_filename);
			return(1);
		}
		if(sscanf(buf2,"%lf",&file_e[i]) != 1) {
			printf("non numeric field on line %d of %s\n", 
			i + 1,my_filename);
			return(1);
		}
		i++;
	} 	
	num_read = i;
	return(0);
}

/*	added by rrt
*/

int
xplot_get_pc(cmd_buf)
char *cmd_buf;
{
	int i, k, c1;
	char c[10];
	char buf[BUFSIZ],*buf2;

	if (fp == NULL) {
		printf("no file open\n");
		return(1);
	}
	if (sscanf(cmd_buf, "%s %d", c, &c1) < 2) {
		printf("%d\n", pc);
		return(0);
	}
	pc = c1;

	fseek(fp, (long) 0, 0);
	i = 0;
	k = 1;

	while ((fgets(buf, BUFSIZ, fp) != NULL) && (i < AR_SIZE)) {
		if (k++ < low_line)
			continue;
		if ((i >= AR_SIZE) || (k > (high_line + 1)))
			break;
		if(iscomment(buf))
			continue;
		if(my_get_col(buf,pc,&buf2)){
			printf("column not found on line %d of %s\n", 
				i + 1,my_filename);
			return(1);
		}
		if(sscanf(buf2,"%lf",&file_p[i]) != 1) {
			printf("non numeric field on line %d of %s\n", 
				i + 1,my_filename);
			return(1);
		}
		i++;
	} 	
	num_read = i;
	return(0);
}

int
xplot_get_xsixty(cmd_buf)
char *cmd_buf;
{
	int i, k, c1;
	char c[10];
	char buf[BUFSIZ],*buf2;

	if (fp == NULL) {
		printf("no file open\n");
		return(1);
	}
	if (sscanf(cmd_buf, "%s %d", c, &c1) < 2) {
		printf("%d\n", xc);
		return(0);
	}
	xc = c1;

	fseek(fp, (long) 0, 0);
	i = 0;
	k = 1;

	while ((fgets(buf, BUFSIZ, fp) != NULL) && (i < AR_SIZE)) {
		if (k++ < low_line) {
			continue;
		}
		if ((i >= AR_SIZE) || (k > (high_line + 1)))
			break;
		if(iscomment(buf))
			continue;
		else {
			if(my_get_col(buf,xc,&buf2)){
				printf("column not found on line %d of %s\n", 
					i + 1,my_filename);
				return(1);
			}
			if(is_hms(buf2) ==0){
				printf("buf2='%s'\n",buf2);
				printf("non hms data on line %d of %s\n",
					i + 1,my_filename);
				return(1);
			}
			file_x[i]=hms2d(buf2);
		}
		i++;
	}
 	num_read = i;
	x_mode=MODE_BABY;
	return(0);
}

int
xplot_get_ysixty(cmd_buf)
char *cmd_buf;
{
	int i, k, c1;
	char c[10];
	char buf[BUFSIZ],*buf2;

	if (fp == NULL) {
		printf("no file open\n");
		return(1);
	}
	if (sscanf(cmd_buf, "%s %d", c, &c1) < 2) {
		printf("%d\n", xc);
		return(0);
	}
	yc = c1;

	fseek(fp, (long) 0, 0);
	i = 0;
	k = 1;

	while ((fgets(buf, BUFSIZ, fp) != NULL) && (i < AR_SIZE)) {
		if (k++ < low_line) {
			continue;
		}
		if ((i >= AR_SIZE) || (k > (high_line + 1)))
			break;
		if(iscomment(buf))
			continue;
		else {
			if(my_get_col(buf,yc,&buf2)){
				printf("column not found on line %d of %s\n", 
					i + 1,my_filename);
				return(1);
			}
			if(is_hms(buf2) ==0){
				printf("non hms data on line %d of %s\n", 
					i + 1,my_filename);
				return(1);
			}
			file_y[i]=hms2d(buf2);
		}
		i++;
	}
 	num_read = i;
	y_mode=MODE_BABY;
	return(0);
}

/*	
 * waits for  
 *     a. the desired number of seconds (if given a number)
 *     b. the data file to change creation time (if given "data")
 *     c. a file full of XPLOT commands to change creation time (otherwise)
 */

int
xplot_wait(cmd_buf)
char	*cmd_buf;
{
	char answer[20], new_cmd[50];
	int seconds;
	struct stat old, new;

	sscanf(cmd_buf,"%*s %s",answer);

	/* case A. sleep for a given number of seconds */
	if (sscanf(answer, "%d", &seconds) == 1) {
		do_flush();
		sleep(seconds);
		return(0);
	}

	/* case B. user supplies "data", so wait until data file is modified */
	if (strcmp(answer, "data") == 0) {
		if (stat(my_filename, &old)<0) {
			printf("stat error when examining file %s - aborting\n",
					my_filename);
			return 1;
		}
		printf("waiting for data file %s to change ...", my_filename);
		fflush(stdout);
		do_flush();
		while(1) {
			stat(my_filename, &new);
			if (old.st_mtime != new.st_mtime) {
				break;
			}
			sleep(1);
		}
		printf("changed\n");
		return(0);
	}

	/* 
	 * case C. wait for given file to be modified, then try to run it,
	 *         assuming it's full of XPLOT commands 
	 *
	 * The string "answer" contains the file's name when we enter this
	 *   code fragment.
	 */
	if (stat(answer, &old) < 0) {
		printf("stat error when examining file %s - aborting\n",
				answer);
		return 1;
	}
	printf("waiting for command file %s to change ...", answer);
	fflush(stdout);
	do_flush();
	while(1) {
		stat(answer, &new);
		if (old.st_mtime != new.st_mtime) {
			break;
		}
		sleep(1);
	}
	/* now run commands from this file */
	sprintf(new_cmd, "run %s", answer);
	xplot_run_cmd(new_cmd);
	return(0);
}

		
/*	added by rrt
*/

int
xplot_errorbar(cmd_buf)
char *cmd_buf;
{
	int K=0;

	if (sscanf(cmd_buf, "%*s %d", &K) < 1) {
		printf("%d\n", K);
		return(0);
	}
	xplot_do_errorbar(K);
	return(0);
}


/* get the limits for the plot - if called without arguments,
	   set limits by the extremes of the data points. put the limits
	   into the variables:
				minx, maxx, miny, maxy
	   in that order.

	if * is in place of a field, then it auto scales
		-added by rrt

	if no args, set limits to extrema and PRINTS OUT THE VALUES  
	    - MWR 11/18/1991
	
	if - is in place of a field then it does nothing 8/10/93 -rrt
*/


int
xplot_get_limits(cmd_buf)
char *cmd_buf;
{
	char  one[20], two[20], three[20], four[20];
	int i,n, done=0;
	double diff,atof();
	
	
	n=sscanf(cmd_buf, "%*s %s %s %s %s", one, two, three, four); 

	if( n<1 || strcmp(one,"*")==0){
		if(num_read<2){
			printf("not enough points\n");
			return 0;
		}
		for (i = 1,minx=file_x[0] , done|=1; i < num_read; i++) 
			if (file_x[i] < minx)
				minx=file_x[i];
	}else if(strcmp(one,"-") !=0)
		minx=atof(one);

	if( n<2 || strcmp(two,"*")==0){
		if(num_read<2){
			printf("not enough points\n");
			return 0;
		}
		for (i = 1,maxx=file_x[0] , done|=1; i < num_read; i++) 
			if (file_x[i] > maxx)
				maxx=file_x[i];
	}else if(strcmp(two,"-") !=0)
		maxx=atof(two);
 
	if( n<3 || strcmp(three,"*")==0){
		if(num_read<2){
			printf("not enough points\n");
			return 0;
		}
		for (i = 1,miny=file_y[0] , done|=2; i < num_read; i++) 
			if (file_y[i] < miny)
				miny=file_y[i];
	}else if(strcmp(three,"-") !=0)
		miny=atof(three);


	if( n<4 || strcmp(four,"*")==0){
		if(num_read<2){
			printf("not enough points\n");
			return 0;
		}
		for (i = 1,maxy=file_y[0] , done|=2; i < num_read; i++) 
			if (file_y[i] > maxy)
				maxy=file_y[i];
	}else if(strcmp(four,"-") !=0)
		maxy=atof(four);
	
	if(done&1){
		diff = (maxx - minx) / 20;
		maxx += diff; 
		minx -= diff; 
	}
	if(done&2){
		diff = (maxy - miny) / 20;
		maxy += diff; 
		miny -= diff; 
	}

	printf("%f %f %f %f\n", minx, maxx, miny, maxy);

	return(0);
}


	/* use only data from between the given two lines in plotting
	   and so forth */

int
xplot_lines(cmd_buf)
char *cmd_buf;
{
	char dummy[10];
	int num, l, h;

	if ((num = sscanf(cmd_buf, "%s %d %d", dummy, &l, &h)) == 1) {
		printf("%d %d\n", low_line, high_line);
	}
	else if (num < 3) {
		printf("usage: lines low high\n");
		return(1);
	}
	else {
		if ((l >= 1) && (l < AR_SIZE))
			low_line = l;
		if ((h >= 1) && (h < AR_SIZE))
			high_line = h;
	}
	return(0);
}


	/* if given two arguments, move the text position to the
	   given place in file (not NDC) coordinates. if given no
	   arguments, print the current text position. if given one,
	   print an error message. 
		the 'draw' command now updates the x_text_position -RRT
*/

int
xplot_position(cmd_buf)
char *cmd_buf;
{
	int num;
	char dummy[10];
	double px, py;

	if ((num = sscanf(cmd_buf, "%s %lf %lf", dummy, &px, &py)) == 1) {
		printf("%f %f\n", x_text_pos, y_text_pos);
	}
	else if (num >= 3) {
		x_text_pos = px;
		y_text_pos = py;
	}
	else {
		printf("usage: position x_pos y_pos\n");
		return(1);
	}
	return(0);
}

	/* take the log (base 10) of the x-coordinates, and use that
	   as the x-coords. for 0 and negative numbers, make the
	   log = BIGNEGNUM */

#define BIGNEGNUM -50

int
xplot_x_log(cmd_buf)
char *cmd_buf;
{
	int i;

	for (i = 0; i < num_read; i++) {
		if (file_x[i] <= 0.) 
			file_x[i] = BIGNEGNUM;	
		else
			file_x[i] = log10(file_x[i]);
	}
	return(0);
}

	/* take the log10 of the y-coords */

int
xplot_y_log(cmd_buf)
char *cmd_buf;
{
	int i;

	for (i = 0; i < num_read; i++) {
		if (file_y[i] <= 0.) 
			file_y[i] = BIGNEGNUM;	
		else
			file_y[i] = log10(file_y[i]);
	}
	return(0);
}

	/* change the x-coords to be '10 to the power of x-coords' - 
	   this will change coords which have been 'logged' back to
	   their original forms (we hope). if the coord is less than
	   or equal to BIGNEGNUM, set the x-coord to zero. */

int
xplot_x_delog(cmd_buf)
char *cmd_buf;
{
	int i;
	double ten = 10.0;

	for (i = 0; i < num_read; i++) {
		if (file_x[i] <= BIGNEGNUM) 
			file_x[i] = 0;	
		else
			file_x[i] = pow(ten, file_x[i]);
	}
	return(0);
}

	/* set the y-coords to '10 to the power of y-coords' */

int
xplot_y_delog(cmd_buf)
char *cmd_buf;
{
	int i;
	double ten = 10.0;

	for (i = 0; i < num_read; i++) {
		if (file_y[i] <= BIGNEGNUM) 
			file_y[i] = 0.0;	
		else
			file_y[i] = pow(ten, file_y[i]);
	}
	return(0);
}

/*	tells whether a data line is to be ignored
		added by rrt
*/

static int
iscomment(in)
char *in;
{
	if((in[0]==';') || (in[0]=='#'))
		return 1;
	else
		return 0;
}


/* a handy command that does several common commands at once:
		data <file>
		xc 1
		yc 2
		limits
		box
		connect
returns a code of 1 if any of the commands had a problem */

int
xplot_plot(cmd_buf)
char *cmd_buf;
{
	char dummy[10], filen[30], buf[50];

	if (sscanf(cmd_buf, "%s %s", dummy, filen) != 2) {
		printf("No file specified\n");
		return(1);
	}
	sprintf(buf, "data %s", filen);
	if (xplot_get_file(buf) != 0)
		return(1);
	if (xplot_get_xc("xc 1") != 0)
		return(1);
	if (xplot_get_yc("yc 2") != 0)
		return(1);
	if (xplot_get_limits("limits") != 0)
		return(1);
	if (xplot_draw_axes("box") != 0)
		return(1);
	if (xplot_connect("connect") != 0)
		return(1);
	return(0);
}
	
/*	
	Vector field command modified 7/28/91-RRT
*/



	/* translate the file_x[] and file_y[] arrays into 
	   x[] and y[] arrays in NDC space, 
	   but don't round to integers in preparation for display
	   yet - do that later so that we don't "wrap around"
	   on the screen due to overflow. */

static void
normalize(void)
{
	int i;

	for (i = 0; i < num_read; i++) {
		world_to_ndc(file_x[i], file_y[i], &(x[i]), &(y[i]));
	}
}
	

	/* return 1 if the given point is inside the window, 0 otherwise */
	/* again, note apparent 'backwards' tests on Y coord due to IBM
	   having (0,0) in the UPPER-left hand corner */

static int
inside(x, y)
double x, y;
{
	if ((x < MINX) || (x > MAXX) || (y > MINY) || (y < MAXY))
		return(0);
	return(1);
}


	/* put a point at each data point */

int
xplot_points(char *cmd_buf)
{
	int i;
	double *px, *py;
	INT16 xy[2];

	normalize();
	for (i = 0, px = x, py = y; i < num_read; i++, px++, py++) {
		if (inside(*px, *py)) {
			xy[0] = *px;
			xy[1] = *py;
			if(file_p[i]>1.0)
				setptype(file_p[i]);	
			draw_marker(xy[0], xy[1]);
		}
	}
	return(0);

}
	/*	converts bizarre  pcolumn syntax 
 		to primitive commands */

static void 
setptype(in)
double in;
{
	int n,s;
	double ex_factor;
	char buf[80];

	ex_factor=10.0*(in-(int)in);
	if(ex_factor>.0099)
		p_size=ex_factor;
		
	n=(int) in/10.;
	s=(int) (in - (double)n * 10.);
	sprintf(buf, "ptype %d %d",n,s);
	xplot_ptype(buf);
}

	/* connect the dots 
		if a second parameter is specified it slows down the connecting
	*/

int
xplot_connect(cmd_buf)
char *cmd_buf;
{
	int i;
	int delay=0;
	char dummy[80];

	sscanf(cmd_buf, "%s %d", dummy, &delay); 
	if(delay < 0 )
		delay=0;

	normalize();
	for (i = 0; i < num_read - 1; i++){
		clip_n_draw(x[i], y[i], x[i+1], y[i+1]);
		if(delay){
			do_flush();
			vista_usleep(delay*10000);
		}
	}

	return(0);
}

	/* draw a histogram style plot */

int
xplot_hist(char *cmd_buf)
{
	int i;
	double newx,oldx;

	normalize();

	oldx=x[0];

	for (i = 0; i < num_read - 1; i++){
		newx=(x[i+1]+x[i])*0.5;
		clip_n_draw(oldx, y[i], newx, y[i]);
		clip_n_draw(newx, y[i], newx, y[i+1]);
		oldx=newx;
	}

	return(0);
}


	/* draw a box around the graphics window area */

static void
box(void)
{
	INT16 corner[10];
	INT16 old_style, new_style;	

	old_style = cur_l_type;
	new_style = SOLID;		/* solid line */

	corner[0] = MINX;	corner[1] = MINY;
	corner[2] = MAXX;	corner[3] = MINY;
	corner[4] = MAXX;	corner[5] = MAXY;
	corner[6] = MINX;	corner[7] = MAXY;
	corner[8] = MINX;	corner[9] = MINY;
	
	draw_poly(5, corner);
	cur_l_type = old_style;
}


	/* place the given string at the current text position - 
	   if it is off-screen, do nothing and let him figure it
	   out. */

int
xplot_label(cmd_buf)
char *cmd_buf;
{
	double dx, dy;
	INT16 px, py;
	char *p;

	/* skip over the word "label" and any blank space to get to the
	   text of the command */
	for (p = cmd_buf + 5; (isspace(*p)); p++)
		;

	/* if no text, just return */
	if (*p == '\0')
		return(0);

	world_to_ndc(x_text_pos, y_text_pos, &dx, &dy);
	if (inside(dx, dy))
		draw_text((px = dx), (py = dy), p);

	return(0);
}

static int
code(x, y)
double x, y;
{

	/* looks backwards due to screen on IBM RT being upside-down.. */

	return(((y < MAXY) << 3) |
	       ((y > MINY) << 2) |
	       ((x > MAXX) << 1) |
	       ((x < MINX)));
}
	
#define SWAP  t = x1; u = y1; x1 = x2; y1 = y2; x2 = t; y2 = u; code1 = code2;

static void
clip_n_draw(x1, y1, x2, y2)
double x1, y1, x2, y2;
{
	double t, u;
	INT16 xy[4];
	int code1, code2;

	while (1) {
		code1 = code(x1, y1);
		code2 = code(x2, y2);
		if (!(code1 | code2)) {
			xy[0] = x1; xy[1] = y1; xy[2] = x2; xy[3] = y2;
			draw_poly(2, xy);
			return;
		}
		else if (code1 & code2)
			return;
		else {

			/* first, make sure that point 1 is outside the window */
			if (!code1) {
				SWAP
			}

			if (code1 & 8) {
				x1 += (x2 - x1) * (MAXY - y1) / (y2 - y1);
				y1 = MAXY;
			}
			else if (code1 & 4) {
				x1 += (x2 - x1) * (MINY - y1) / (y2 - y1);
				y1 = MINY;
			}
			else if (code1 & 2) {
				y1 += (y2 - y1) * (MAXX - x1) / (x2 - x1);
				x1 = MAXX;
			}
			else if (code1 & 1) {
				y1 += (y2 - y1) * (MINX - x1) / (x2 - x1);
				x1 = MINX;
			}
		}	
	}
}

	/* put a point marker at the current 'position' coordinates */ 

int
xplot_dot(cmd_buf)
char *cmd_buf;
{
	double nx, ny;
	INT16 x, y;

	world_to_ndc(x_text_pos, y_text_pos, &nx, &ny);
	if (inside(nx, ny)) {
		x = nx;
		y = ny;
		draw_marker(x, y);
	}
	return(0);
}

	/* draw a line from the current 'position' coordinates to the
	   given coordinates on the command line 
		after drawing.. the x_text_pos and y_text_pos are updated-rrt
	*/

int
xplot_draw(cmd_buf)
char *cmd_buf;
{
	double wx, wy, x1, y1, x2, y2;
	char dummy[10];

	if (sscanf(cmd_buf, "%s %lf %lf", dummy, &wx, &wy) < 3) {
		printf("usage: draw x-coord y-coord\n");
		return(1);
	}
	world_to_ndc(x_text_pos, y_text_pos, &x1, &y1);
	world_to_ndc(wx, wy, &x2, &y2);
	clip_n_draw(x1, y1, x2, y2);
	x_text_pos=wx;
	y_text_pos=wy;
	return(0);
}
	
/*	vector field drawing routine - rrt
		vector tail is in x,y
		vector angle in degrees CCW from right in ecol
		vector length in drawing units in pcol
	p_size determines expansion
*/

#define PF  (M_PI/180.)	/* conversion factor from degrees to radians */
#define HEAD 10.		/* head size */
#define EXTRA  20.		/* angle off the line for the head */

int
xplot_vfield(char *cmd_buf)
{
	int i;
	double x1,x2,y1,y2,x3,y3,x4,y4;

	for (i = 0 ; i < num_read; i++) {
		world_to_ndc(file_x[i], file_y[i], &x1, &y1);
		if (inside(x1, y1)) {
			x2=x1 + p_size * file_p[i]*cos(PF*file_e[i]);
			y2=y1 - p_size * file_p[i]*sin(PF*file_e[i]);
			clip_n_draw(x1,y1, x2, y2);
			x3=x2 + p_size *HEAD*cos(PF*(file_e[i]+180.+ EXTRA));
			y3=y2 - p_size *HEAD*sin(PF*(file_e[i]+180.+ EXTRA));
			clip_n_draw(x2,y2, x3, y3);
			x4=x2 + p_size *HEAD*cos(PF*(file_e[i]+180.- EXTRA));
			y4=y2 - p_size *HEAD*sin(PF*(file_e[i]+180.- EXTRA));
			clip_n_draw(x2,y2, x4, y4);
		}
	}
	return(0);
}

int
xplot_do_errorbar(K)
int K;
{
	int i;
	double x1,x2,y1,y2,fx1,fy1,fx2,fy2,rot;

	for (i = 0 ; i < num_read; i++) {
		fx1=file_x[i];
		fy1=file_y[i];
		world_to_ndc(fx1, fy1, &x1, &y1);
		if (inside(x1, y1)) {
			rot=90.*(K-1);
			fx2=fx1 + file_e[i]*cos(PF*rot);
			fy2=fy1 + file_e[i]*sin(PF*rot);
			world_to_ndc( fx2, fy2, &x2, &y2);
			clip_n_draw(x1,y1, x2, y2);
		}
	}
	return(0);
}


/*	
	Returns the value of col from buf
	returns 1 if an error occurs
*/

static int
my_get_col(buf, col, buf2)
char *buf;
int col;
char **buf2;
{
	int i = 0;

	while (--col) {
		while (isspace(buf[i])) {	/*skip leading blanks */
			i++;		
			if (buf[i] == '\0')		
				return 1;
		}
		while (!isspace(buf[i])) { /*find next whitespace */
			i++;		
			if (buf[i] == '\0')		
				return 1;
		}
	}
	while (isspace(buf[i])) {	/*skip leading blanks */
		i++;		
		if (buf[i] == '\0')		
			return 1;
	}
	*buf2=&buf[i];
	while (!isspace(buf[i])) { /*find next whitespace */
			if (buf[i] == '\0')		
				break;
			i++;		
		}
	buf[i]=0;			/* zero terminate */
	return 0;
}




/****************************************************************************/

/*	additional math routines 
	Author: R.R. Treffers
*/


#define NBUF		80
#define	SMALL		1.0e-8			/* small floating number		*/

static int sgn( /* double */ );
static void blanktozero( /* char * */);
	/* because the damn sprintf doesn't work */


char *d2hms(x,mode)		/* converts double to babylonian notation */
double x;		
int mode;
{
	int sign,x1,x2;
	static char pminus[]={'-',' ','+'};
	static char ra_buf[15],dec_buf[15],time_buf[15],goof[]={"**********"};
	double q,x3;
	char *rval=NULL;

	if(fabs(x)>99.)
		return(goof);

	sign=sgn(x);
	x1=(int)(x=fabs(x)+SMALL);
	x2=(int)(q=(x-(double)x1)*60.+SMALL);
	x3=((double)q-x2)*60.+SMALL;

	if(mode==BABY_MODE_RA){
		if(x3>59.95){
			x3=0.;
			x2++;
			if(x2>59){
				x2=0;
				x1++;
			}
		}	
		sprintf(ra_buf,"%2d %2.2d %04.1f",x1,x2,x3);
		rval=ra_buf;
	}else

	if(mode==BABY_MODE_DEC){
		if(x3>59.5){
			x3=0.;
			x2++;
			if(x2>59){
				x2=0;
				x1++;
			}
		}	
		sprintf(dec_buf,"%1c%2d %2.2d %02.0f",pminus[sign+1],x1,x2,x3);
		rval=dec_buf;
	}else
	
	if(mode==BABY_MODE_RA_NO){
		if(x3>59.95){
			x3=0.;
			x2++;
			if(x2>59){
				x2=0;
				x1++;
			}
		}	
		sprintf(ra_buf,"%2.2d:%2.2d:%04.1f",x1,x2,x3);
		blanktozero(ra_buf);
		rval=ra_buf;	
	}else
	
	if(mode==BABY_MODE_DEC_NO){
		if(x3>59.5){
			x3=0.;
			x2++;
			if(x2>59){
				x2=0;
				x1++;
			}
		}	
		sprintf(dec_buf,"%1c%2.2d:%2.2d:%02.0f",pminus[sign+1],x1,x2,x3);
		blanktozero(dec_buf);
		rval=dec_buf;
	}else
	if(mode==BABY_MODE_TIME){
		if(x3>59.5){
			x3=0.;
			x2++;
			if(x2>59){
				x2=0;
				x1++;
			}
		}	
		sprintf(time_buf,"%2d %2.2d %02.0f",x1,x2,x3);
		rval=time_buf;
	}else
	if(mode==BABY_MODE_TIME_NO){
		if(x3>59.5){
			x3=0.;
			x2++;
			if(x2>59){
				x2=0;
				x1++;
			}
		}	
		sprintf(time_buf,"%2d:%2.2d:%02.0f",x1,x2,x3);
		blanktozero(time_buf);
		rval=time_buf;
	}else
	if(mode==BABY_MODE_DEC_NO2){
		if(x3>59.95){
			x3=0.;
			x2++;
			if(x2>59){
				x2=0;
				x1++;
			}
		}	
		sprintf(dec_buf,"%1c%2.2d:%2.2d:%04.1f",pminus[sign+1],x1,x2,x3);
		blanktozero(dec_buf);
		rval=dec_buf;	
	}else		
		printf("illegal case in d2hms\n");

	return rval;
}

double hms2d(in)		/* converts babylonian string to double */
char *in;
{
	int sign_count=0,got_period=0,nfield;
	double deg=0.,min=0.,sec=0.,val,sign=1.;
	char *p,copy[NBUF],*start;

	strcpy(copy,in);

	for(p=start=copy; *p ; p++){
		switch(*p){
			case ':':				/* convert colons to spaces */
				*p=' ';
				break;
			case '-':				/* search for - sign  and change sign */
				sign=(-1.);
				start=p+1;					/* skip over for sscanf function */	
				if(sign_count++==2)
					printf("improperly formed hms string\n");
				break;
			case '+':					
				sign=1.;
				start=p+1;
				if(sign_count++==2)
					printf("improperly formed hms string\n");
				break;
			case '.':
				got_period=1;
			case ' ':
				break;
			default:
				if(!isdigit(*p)){
					printf("illegal char in hms string '%s'\n", in);
				}
		}
	}
	
	nfield=sscanf(start,"%lf %lf %lf",&deg,&min,&sec);
	if((fabs(deg)>360.) || (min>59.) || (min<0.) || (sec>60.) || (sec<0.))
		printf("improperly formed hms string\n");

	if((nfield==1) && got_period){		/* this is the old method of conversion */
		int x1,x2;					/* which will be made obsolete */
		double x3,q,x;

		x=atof(start);
		x1=(int)(x+.1);
		q=(x-(double)x1)*100.;
		x2=(int)(q+.1);
		x3=(q-(double)x2)*100.;
		val=(double)x1+(double)x2/60.+x3/3600.;
	}else
		val=deg+min/60.+sec/3600.;
	return(sign*val);
}

static int
is_hms(in)		/* returns 1 if valid babylonian string */ 
char *in;
{
	int sign_count=0, got_period=0, nfield;
	double deg=0.0, min=0.0, sec=0.0, sign=1.0;
	char *p, copy[NBUF], *start;

	if(strlen(in) == 0)		/* rather minimalist */
		return 0;

	strcpy(copy,in);

	for(p=start=copy     ; *p ; p++){
		switch(*p){
			case ':':				/* convert colons to spaces */
				*p=' ';
				break;
			case '-':				/* search for - sign  and change sign */
				sign=(-1.);
				start=p+1;					/* skip over for sscanf function */	
				if(sign_count++==2)
					return 0;
				break;
			case '+':					
				sign=1.;
				start=p+1;
				if(sign_count++==2)
					return 0;
				break;
			case '.':
				got_period=1;
			case ' ':
				break;
			default:
				if(!isdigit(*p)){
					return 0;
				}
		}
	}
	
	nfield=sscanf(start,"%lf %lf %lf",&deg,&min,&sec);
	if((fabs(deg)>360.) || (min>59.) || (min<0.) || (sec>60.) || (sec<0.))
		return 0;

	return 1;
}



static int sgn(x)			/* signum function */
double x;		
{
	if(x>=0.)
		return(1);
	return(-1);
}

static void blanktozero(in)
char *in;
{
	int i;

	for(i=0; in[i]; i++)
		if(in[i]==' ')
			in[i]='0';
}





/****************************************************************************/
/*                                                                          */
/*                                                                          */
/*               X Windows routines follow ....                             */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/****                                                                    ****/


#ifndef M_2PI
#define M_2PI           6.2831853		/* two times pi */
#endif

#define FONT_NAME   "9x15"

#define NORMAL          0
#define INVERTED        1

#define BORDER_WIDTH    4
#define LINE_WIDTH      1
#define LINE_STYLE      LineSolid
#define CAP_STYLE       CapButt
#define JOIN_STYLE      JoinMiter
#define TEXT_CHAR_BDR   5

#define MESSAGE_LENGTH	20	/* size of Client message buffer */

static Display *display;
static int screen;
static XSizeHints size_hints;
static GC graphics_gc[2],  copy_gc, invert_gc;
static Cursor box_cursor;
static XGCValues values;
static XFontStruct *fontinfo;
static Window graphics_window;
static XWindowAttributes xwa;
static int graphics_width, graphics_height;
static int char_width, char_height;
static Pixmap  graphics_pixmap;
static int display_mode;
static unsigned long serial;	/* used by data transmission */
static Atom xplotatom;		/* used for data transmission */
static char atomname[]="xplotatom" ;
static XPoint XVert[AR_SIZE];
static XPoint p_vert[P_MAXVERT+2];
static int p_nvert;


static void
initXstuff(void)
{

	if ((display = XOpenDisplay(NULL)) == NULL)
		die("Sorry, can't open display - are you in X-Windows?");
	screen = DefaultScreen(display);
	values.function = GXcopy;
	copy_gc = XCreateGC(display, RootWindow(display, screen), GCFunction,
			&values);
	box_cursor = XCreateFontCursor(display, XC_dotbox);
	values.function = GXcopyInverted;
	invert_gc = XCreateGC(display, RootWindow(display, screen), GCFunction,
			&values);
	
	if ((fontinfo = XLoadQueryFont(display, FONT_NAME)) == NULL)
		die("can't open font");
	char_height = fontinfo->max_bounds.ascent + fontinfo->max_bounds.descent;
	char_width = get_char_width();

	graphics_width = GRAPHICS_WIDTH;
	graphics_height = GRAPHICS_HEIGHT;
	if ((graphics_window = XCreateSimpleWindow(display, RootWindow(display,
			screen), GRAPHICS_XOFFSET, GRAPHICS_YOFFSET, graphics_width,
			graphics_height, BORDER_WIDTH, WhitePixel(display, screen),
			BlackPixel(display, screen))) < 0)
		die("error in XCreateSimpleWindow for graphics window");

	size_hints.flags = USPosition | USSize;
	size_hints.width = graphics_width;
	size_hints.height = graphics_height;
	size_hints.x = GRAPHICS_XOFFSET;
	size_hints.y = GRAPHICS_YOFFSET;
	size_hints.min_width = graphics_width;
	size_hints.min_height = graphics_height;
	if (XSetStandardProperties(display, graphics_window, "xplot", 
			"xplot", (Pixmap) NULL, (char **) NULL, 0, &size_hints) < 0)
		die("error in XSetStandardProperties");
	if (XSelectInput(display,
				graphics_window,
				ExposureMask | ButtonPressMask )
	 < 0)
		die("error in XSelectInput");
	get_norm_GC(graphics_window, &graphics_gc[NORMAL], fontinfo);
	get_inverted_GC(graphics_window, &graphics_gc[INVERTED], fontinfo);
	XDefineCursor(display, graphics_window, box_cursor);
	XFillRectangle(display, graphics_window, graphics_gc[!display_mode],
			0, 0, graphics_width, graphics_height);

	XMapWindow(display, graphics_window);
	XFillRectangle(display, graphics_window, graphics_gc[!display_mode],
			0, 0, graphics_width, graphics_height);
	if (XGetWindowAttributes(display, graphics_window, &xwa) < 0)
		die("error in XGetWindowAttributes\n");
	if ((graphics_pixmap = XCreatePixmap(display, graphics_window, 
			graphics_width, graphics_height, xwa.depth)) < 0)
		die("can't allocate for graphics pixmap");
	XCopyArea(display, graphics_window, graphics_pixmap, copy_gc, 0, 0, 
			graphics_width, graphics_height, 0, 0);

	xplotatom=XInternAtom(display, atomname, False);
	display_mode = INVERTED;
}

/*	sends the string to the parent and
	notifies
*/

void
xplot_transmit(s)
char *s;
{
	int i, j=0 ,rval;
	XEvent event;
	char last;

	event.type=ClientMessage;
	event.xany.serial=serial;
	event.xany.send_event=True;
	event.xany.display=display;
	event.xclient.window=graphics_window;		/* i don't know */
	event.xclient.message_type=xplotatom;	
	event.xclient.format=8;						/* denotes byte data */

	do{
		for(i=0; i<MESSAGE_LENGTH; i++, j++){ 
			event.xclient.data.b[i]=last=s[j];
			if(s[j] == 0)
				break;
		}
		rval=XSendEvent(display, 
				graphics_window,
				True,
				ExposureMask | ButtonPressMask,
				&event);
	}while(last);
	XFlush(display);
}
		
static void
endXstuff(void)
{
	xplot_clear();
	XUnloadFont(display, fontinfo->fid);
	XFreeGC(display, copy_gc);
	XFreeGC(display, invert_gc);
	XFreeGC(display, graphics_gc[NORMAL]);
	XFreeGC(display, graphics_gc[INVERTED]);
	XFreePixmap(display, graphics_pixmap);
	XCloseDisplay(display);
	XFlush(display);
}


/*	proceses the X-window events, will return
	buffer from gets() which has been forked 

	the "break_flag" argument controls whether we return after
	processing one event:
	  XPLOT_EVENT_LOOP         do not return, stay in loop
	  XPLOT_EVENT_BUTTON       break if the user typed a mouse button
	
*/

void
xplot_event(s, break_flag)
char *s;
int break_flag;
{
	XEvent event;
	int i, j=0, done=0;
	
	while( !done ){
		XNextEvent(display, &event);
		serial=event.xany.serial;
		switch(event.type){
			case Expose:
				XCopyArea(display, graphics_pixmap, graphics_window, 
					copy_gc, 0, 0, graphics_width, graphics_height, 0, 0);
				break;
			case ButtonPress:
				localcursor(event.xbutton.x,event.xbutton.y);
				if (break_flag == XPLOT_EVENT_BUTTON) {
					done = 1;
				}
				break;
			case ClientMessage:
				if(event.xclient.message_type==xplotatom){
					for(i=0; i<MESSAGE_LENGTH; i++,j++){
						s[j]=event.xclient.data.b[i];
						if(s[j]==0){
							done=1;
							break;
						}
					}
				}
				break;
		}
	}
}


	/* clear the graphics window */

int
xplot_clear(void)
{
	adobe_erase();
	XFillRectangle(display, graphics_window, graphics_gc[!display_mode],
			0, 0, graphics_width, graphics_height);
	XCopyArea(display, graphics_window, graphics_pixmap, copy_gc,
			0, 0, graphics_width, graphics_height, 0, 0);
	do_flush();

	return(0);
}

	/* return the width of the widest character in the font (in pixels) */

static int
get_char_width(void)
{
	char buf[10];
	int width, i, max;
	
	max = 0;
	buf[1] = '\0';
	for (i = 0; i < 255; i++) {
		buf[0] = i;
		width = XTextWidth(fontinfo, buf, 1);
		if (width > max)
			max = width;
	}
	return(max);
}
		
	/* reverse the graphics screen from black-on-white to white-on-black,
	   or vice versa */
	/* Actually, although this works on monochrome displays, on color
	   displays, it destroys the current window contents. 
			MWR 7/28/96	*/

int
xplot_reverse(char *cmd_buf)
{

	if (display_mode == NORMAL) {
		display_mode = INVERTED;
	}
	else {
		display_mode = NORMAL;
	}

	XCopyArea(display, graphics_window, graphics_window, invert_gc,
			0, 0, graphics_width, graphics_height, 0, 0);

	XCopyArea(display, graphics_window, graphics_pixmap, copy_gc,
			0, 0, graphics_width, graphics_height, 0, 0);
	return(0);
}	

	/* draw a polygon connecting the given vertices */


static void
draw_poly(n, pts)
int n;
INT16 pts[];
{
	int i, j;
	int dash_offset, dash_num;
	char dash_list[10];

	j = 0;
	for (i = 0; i < n; i++) {
		XVert[i].x = pts[j++];
		XVert[i].y = pts[j++];
	}
	if ((cur_l_type == DOTTED) || (cur_l_type == DASH_DOT) ||
		(cur_l_type == DASHED)) {
		dash_offset = 0;
		if (cur_l_type == DOTTED) {
			dash_list[0] = 1;
			dash_list[1] = 3;
			dash_num = 2;
		} 
		else if (cur_l_type == DASHED) {
			dash_list[0] = 4;
			dash_list[1] = 4;
			dash_num = 2;;
		}
		else {
			dash_list[0] = 4;
			dash_list[1] = 3;
			dash_list[2] = 1;
			dash_list[3] = 3;
			dash_num = 4;
		}
		XSetDashes(display, graphics_gc[NORMAL], dash_offset, dash_list,
				dash_num);
		XSetDashes(display, graphics_gc[INVERTED], dash_offset, dash_list,
				dash_num);
		XSetLineAttributes(display, graphics_gc[NORMAL], LINE_WIDTH,
				LineDoubleDash, CAP_STYLE, JOIN_STYLE);
		XSetLineAttributes(display, graphics_gc[INVERTED], LINE_WIDTH,
				LineDoubleDash, CAP_STYLE, JOIN_STYLE);
	}
	else {					/* set line type to SOLID */
		XSetLineAttributes(display, graphics_gc[NORMAL], LINE_WIDTH,
				LineSolid, CAP_STYLE, JOIN_STYLE);
		XSetLineAttributes(display, graphics_gc[INVERTED], LINE_WIDTH,
				LineSolid, CAP_STYLE, JOIN_STYLE);
	}

	XDrawLines(display, graphics_window, graphics_gc[display_mode], 
			XVert, n, CoordModeOrigin);
	adobe_drawpoly(n,(XPOINT *) XVert,ADOBE_ABSOLUTE);
}

	/* place the given text string at the given position in the
	   graphics window */

static void
draw_text(x, y, s)
INT16 x, y;
char *s;
{
	int x1, y1;

	x1 = x;
	y1 = y + fontinfo->max_bounds.ascent;

	XDrawImageString(display, graphics_window, graphics_gc[display_mode],
			x1, y1, s, strlen(s));
	adobe_drawtext(x1,y1,s);
}	


	/* all the stuff in here just draws different kinds of point markers */
	   

	/* set up the list of vertices which will be the point marker 
	   for all subsequent 'points' commands (until changed by another
	   'ptype' command) */
	/* with no args, prints the current type */

int
xplot_ptype(cmd_buf)
char *cmd_buf;
{
	int i, type, fill, num;
	char dummy[7];
	double pa, dpa;
	int newx, newy, oldx, oldy;

	fill = cur_p_fill;
	if ((num = sscanf(cmd_buf, "%s %d %d", dummy, &type, &fill)) < 2) {
		printf("%d %d\n", cur_p_type, cur_p_fill);
		return(0);
	}
	
	if (type < 2) {
		type = 1;
		p_nvert = 1;
		p_vert[0].x = 0;
		p_vert[0].y = 0;
	}
	else {
		if (type > P_MAXVERT)
			type = P_MAXVERT;
		p_nvert = type + 1;
		dpa = M_2PI/type;
		oldx = 0;
		oldy = 0;
		for (i = 0, pa = 0.0; pa <= M_2PI; i++, pa += dpa) {
			newx = P_SIZE * p_size * cos(pa);
			newy = P_SIZE * p_size * sin(pa);
			p_vert[i].x = (newx - oldx);
			p_vert[i].y = (newy - oldy);
			oldx = newx;
			oldy = newy;
		}
	}
	cur_p_type = type;
	cur_p_fill = fill;
	return(0);
}

	/* actually do the point-marker drawing */

static void
draw_marker(x, y)
INT16 x, y;
{
	p_vert[0].x += x;
	p_vert[0].y += y;

	if (cur_p_type == 1) {
		XDrawPoint(display, graphics_window, graphics_gc[display_mode],
				p_vert[0].x, p_vert[0].y);
		adobe_drawpoly(1,(XPOINT *) p_vert,ADOBE_ABSOLUTE);
	}
	else {
		XDrawLines(display, graphics_window, graphics_gc[display_mode],
				p_vert, p_nvert, CoordModePrevious);
		if (cur_p_fill > 0){
			XFillPolygon(display, graphics_window, graphics_gc[display_mode],
					p_vert, p_nvert, Nonconvex, CoordModePrevious);
			adobe_drawpoly(p_nvert,(XPOINT *) p_vert,
				ADOBE_PREVIOUS | ADOBE_FILL);
		}

		adobe_drawpoly(p_nvert,(XPOINT *) p_vert,ADOBE_PREVIOUS);
			
	}

	p_vert[0].x -= x;
	p_vert[0].y -= y;
}


	/* grab the mouse, and if the user touches
		left button:   display current coords, set 'position'  
		right button:  quit this routine
			formatting changed 3/3/91 -rrt
		3/7/91- cursor removed as well as x_text_pos update rrt
	*/

int
xplot_cursor(char *cmd_buf)
{
	printf("Cursor command is redundant... just press left button\n");
	return 0;
}

static void
localcursor(x,y)
int x,y;
{
	double  wx, wy;	
	char xformat[10],yformat[10];

	ndc_to_world((double) x, (double) y, &wx, &wy);
	cursor_xpos = wx;
	cursor_ypos = wy;

	printf("position is now "); 
	if(x_mode == MODE_NUMERIC){
		find_digits(minx, maxx, xformat);
		printf(xformat,wx);
	}else
		printf("%s",d2hms(wx,BABY_MODE_DEC_NO));

	printf(" ");

	if(y_mode == MODE_NUMERIC){
		find_digits(miny, maxy, yformat);
		printf(yformat,wy);
	}else
		printf("%s",d2hms(wy,BABY_MODE_DEC_NO));
	printf("\n");

}

   /* 
    * read the current values of the variables "cursor_xpos" and "cursor_ypos",
    * which are set whenever the user presses the cursor button
    */

void
xplot_get_cursor(x, y)
double *x, *y;
{
   *x = cursor_xpos;
   *y = cursor_ypos;
}





	/* get a normal graphics context for the given window */

static void
get_norm_GC(window, gc, fontinfo)
Window window;
GC *gc;
XFontStruct *fontinfo;
{
	unsigned long valuemask = GCFunction;
	XGCValues values;
	unsigned int line_width = LINE_WIDTH;
	int line_style = LINE_STYLE;
	int cap_style = CAP_STYLE;
	int join_style = JOIN_STYLE;
	int dash_offset = 0;
	static char dash_list[] = { 12, 24 };
	int list_length = 2;

	values.function = GXcopy;
	*gc = XCreateGC(display, window, valuemask, &values);
	XSetFont(display, *gc, fontinfo->fid);
	XSetForeground(display, *gc, WhitePixel(display, screen));
	XSetBackground(display, *gc, BlackPixel(display, screen));
	XSetLineAttributes(display, *gc, line_width, line_style, cap_style,
			join_style);
	XSetDashes(display, *gc, dash_offset, dash_list, list_length);
}
	
	/* get a inverted graphics context for the given window */

static void
get_inverted_GC(window, gc, fontinfo)
Window window;
GC *gc;
XFontStruct *fontinfo;
{
	unsigned long valuemask = GCFunction;
	XGCValues values;
	unsigned int line_width = LINE_WIDTH;
	int line_style = LINE_STYLE;
	int cap_style = CAP_STYLE;
	int join_style = JOIN_STYLE;
	int dash_offset = 0;
	static char dash_list[] = { 12, 24 };
	int list_length = 2;

	values.function = GXcopy;
	*gc = XCreateGC(display, window, valuemask, &values);
	XSetFont(display, *gc, fontinfo->fid);
	XSetForeground(display, *gc, BlackPixel(display, screen));
	XSetBackground(display, *gc, WhitePixel(display, screen));
	XSetLineAttributes(display, *gc, line_width, line_style, cap_style,
			join_style);
	XSetDashes(display, *gc, dash_offset, dash_list, list_length);
}
	
	/* save the graphics and text window contents into their pixmaps */

static void
savepixmaps(void)
{
	XCopyArea(display, graphics_window, graphics_pixmap, copy_gc, 0, 0, 
			graphics_width, graphics_height, 0, 0);
}

static void do_flush()
{
	XFlush(display);
}





